@@ -4,9 +4,10 @@
 #define __sepol_errno_h__
 
 #include <errno.h>
-#include <sys/cdefs.h>
 
-__BEGIN_DECLS
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 #define SEPOL_OK             0
 
@@ -25,5 +26,8 @@ __BEGIN_DECLS
 #define SEPOL_EEXIST         -EEXIST
 #define SEPOL_ENOENT         -ENOENT
 
-__END_DECLS
+#ifdef __cplusplus
+}
+#endif
+
 #endif
